@@ -14,40 +14,48 @@ module Distribution.Client.List (
   ) where
 
 import Distribution.Package
-         ( PackageName(..), Package(..), packageName, packageVersion
-         , Dependency(..), simplifyDependency )
+         ( PackageName, Package(..), packageName
+         , packageVersion, UnitId )
+import Distribution.Types.Dependency
+import Distribution.Types.UnqualComponentName
 import Distribution.ModuleName (ModuleName)
 import Distribution.License (License)
 import qualified Distribution.InstalledPackageInfo as Installed
 import qualified Distribution.PackageDescription   as Source
 import Distribution.PackageDescription
-         ( Flag(..), FlagName(..) )
+         ( Flag(..), unFlagName )
 import Distribution.PackageDescription.Configuration
          ( flattenPackageDescription )
+import Distribution.Pretty (pretty)
 
 import Distribution.Simple.Compiler
         ( Compiler, PackageDBStack )
-import Distribution.Simple.Program (ProgramConfiguration)
+import Distribution.Simple.Program (ProgramDb)
 import Distribution.Simple.Utils
-        ( equating, comparing, die, notice )
+        ( equating, comparing, die', notice )
 import Distribution.Simple.Setup (fromFlag)
+import Distribution.Simple.PackageIndex (InstalledPackageIndex)
 import qualified Distribution.Simple.PackageIndex as InstalledPackageIndex
-import qualified Distribution.Client.PackageIndex as PackageIndex
 import Distribution.Version
-         ( Version(..), VersionRange, withinRange, anyVersion
+         ( Version, mkVersion, versionNumbers, VersionRange, withinRange, anyVersion
          , intersectVersionRanges, simplifyVersionRange )
 import Distribution.Verbosity (Verbosity)
-import Distribution.Text
+import Distribution.Deprecated.Text
          ( Text(disp), display )
 
+import qualified Distribution.SPDX as SPDX
+
+import           Distribution.Solver.Types.PackageConstraint
+import qualified Distribution.Solver.Types.PackageIndex as PackageIndex
+import           Distribution.Solver.Types.SourcePackage
+
 import Distribution.Client.Types
-         ( SourcePackage(..), Repo, SourcePackageDb(..) )
-import Distribution.Client.Dependency.Types
-         ( PackageConstraint(..), ExtDependency(..) )
+         ( SourcePackageDb(..), PackageSpecifier(..), UnresolvedSourcePackage )
 import Distribution.Client.Targets
-         ( UserTarget, resolveUserTargets, PackageSpecifier(..) )
+         ( UserTarget, resolveUserTargets )
 import Distribution.Client.Setup
-         ( GlobalFlags(..), ListFlags(..), InfoFlags(..) )
+         ( GlobalFlags(..), ListFlags(..), InfoFlags(..)
+         , RepoContext(..) )
 import Distribution.Client.Utils
          ( mergeBy, MergeResult(..) )
 import Distribution.Client.IndexUtils as IndexUtils
@@ -58,7 +66,7 @@ import Distribution.Client.FetchUtils
 import Data.List
          ( sortBy, groupBy, sort, nub, intersperse, maximumBy, partition )
 import Data.Maybe
-         ( listToMaybe, fromJust, fromMaybe, isJust )
+         ( listToMaybe, fromJust, fromMaybe, isJust, maybeToList )
 import qualified Data.Map as Map
 import Data.Tree as Tree
 import Control.Monad
@@ -69,32 +77,49 @@ import Text.PrettyPrint as Disp
 import System.Directory
          ( doesDirectoryExist )
 
+import Distribution.Utils.ShortText (ShortText)
+import qualified Distribution.Utils.ShortText as ShortText
+
 
 -- | Return a list of packages matching given search strings.
 getPkgList :: Verbosity
            -> PackageDBStack
-           -> [Repo]
+           -> RepoContext
            -> Compiler
-           -> ProgramConfiguration
+           -> ProgramDb
            -> ListFlags
            -> [String]
            -> IO [PackageDisplayInfo]
-getPkgList verbosity packageDBs repos comp conf listFlags pats = do
-    installedPkgIndex <- getInstalledPackages verbosity comp packageDBs conf
-    sourcePkgDb       <- getSourcePackages    verbosity repos
+getPkgList verbosity packageDBs repoCtxt comp progdb listFlags pats = do
+    installedPkgIndex <- getInstalledPackages verbosity comp packageDBs progdb
+    sourcePkgDb       <- getSourcePackages verbosity repoCtxt
     let sourcePkgIndex = packageIndex sourcePkgDb
         prefs name = fromMaybe anyVersion
                        (Map.lookup name (packagePreferences sourcePkgDb))
 
-        pkgsInfo :: [(PackageName, [Installed.InstalledPackageInfo], [SourcePackage])]
+        pkgsInfo ::
+          [(PackageName, [Installed.InstalledPackageInfo], [UnresolvedSourcePackage])]
         pkgsInfo
             -- gather info for all packages
-          | null pats = mergePackages (InstalledPackageIndex.allPackages installedPkgIndex)
-                                      (         PackageIndex.allPackages sourcePkgIndex)
+          | null pats = mergePackages
+                        (InstalledPackageIndex.allPackages installedPkgIndex)
+                        (         PackageIndex.allPackages sourcePkgIndex)
 
             -- gather info for packages matching search term
-          | otherwise = mergePackages (matchingPackages InstalledPackageIndex.searchByNameSubstring installedPkgIndex)
-                                      (matchingPackages (\ idx n -> concatMap snd (PackageIndex.searchByNameSubstring idx n)) sourcePkgIndex)
+          | otherwise = pkgsInfoMatching
+
+        pkgsInfoMatching ::
+          [(PackageName, [Installed.InstalledPackageInfo], [UnresolvedSourcePackage])]
+        pkgsInfoMatching =
+          let matchingInstalled = matchingPackages
+                                  InstalledPackageIndex.searchByNameSubstring
+                                  installedPkgIndex
+              matchingSource   = matchingPackages
+                                 (\ idx n ->
+                                   concatMap snd
+                                   (PackageIndex.searchByNameSubstring idx n))
+                                 sourcePkgIndex
+          in mergePackages matchingInstalled matchingSource
 
         matches :: [PackageDisplayInfo]
         matches = [ mergePackageInfo pref
@@ -115,14 +140,14 @@ getPkgList verbosity packageDBs repos comp conf listFlags pats = do
 -- | Show information about packages.
 list :: Verbosity
      -> PackageDBStack
-     -> [Repo]
+     -> RepoContext
      -> Compiler
-     -> ProgramConfiguration
+     -> ProgramDb
      -> ListFlags
      -> [String]
      -> IO ()
-list verbosity packageDBs repos comp conf listFlags pats = do
-    matches <- getPkgList verbosity packageDBs repos comp conf listFlags pats
+list verbosity packageDBs repos comp progdb listFlags pats = do
+    matches <- getPkgList verbosity packageDBs repos comp progdb listFlags pats
 
     if simpleOutput
       then putStr $ unlines
@@ -145,9 +170,9 @@ list verbosity packageDBs repos comp conf listFlags pats = do
 
 info :: Verbosity
      -> PackageDBStack
-     -> [Repo]
+     -> RepoContext
      -> Compiler
-     -> ProgramConfiguration
+     -> ProgramDb
      -> GlobalFlags
      -> InfoFlags
      -> [UserTarget]
@@ -155,11 +180,11 @@ info :: Verbosity
 info verbosity _ _ _ _ _ _ [] =
     notice verbosity "No packages requested. Nothing to do."
 
-info verbosity packageDBs repos comp conf
+info verbosity packageDBs repoCtxt comp progdb
      globalFlags _listFlags userTargets = do
 
-    installedPkgIndex <- getInstalledPackages verbosity comp packageDBs conf
-    sourcePkgDb   <- getSourcePackages    verbosity repos
+    installedPkgIndex <- getInstalledPackages verbosity comp packageDBs progdb
+    sourcePkgDb       <- getSourcePackages verbosity repoCtxt
     let sourcePkgIndex = packageIndex sourcePkgDb
         prefs name = fromMaybe anyVersion
                        (Map.lookup name (packagePreferences sourcePkgDb))
@@ -168,14 +193,16 @@ info verbosity packageDBs repos comp conf
         -- just available source packages, so we must resolve targets using
         -- the combination of installed and source packages.
     let sourcePkgs' = PackageIndex.fromList
-                    $ map packageId (InstalledPackageIndex.allPackages installedPkgIndex)
-                   ++ map packageId (         PackageIndex.allPackages sourcePkgIndex)
-    pkgSpecifiers <- resolveUserTargets verbosity
+                    $ map packageId
+                      (InstalledPackageIndex.allPackages installedPkgIndex)
+                   ++ map packageId
+                      (PackageIndex.allPackages sourcePkgIndex)
+    pkgSpecifiers <- resolveUserTargets verbosity repoCtxt
                        (fromFlag $ globalWorldFile globalFlags)
                        sourcePkgs' userTargets
 
     pkgsinfo      <- sequence
-                       [ do pkginfo <- either die return $
+                       [ do pkginfo <- either (die' verbosity) return $
                                          gatherPkgInfo prefs
                                            installedPkgIndex sourcePkgIndex
                                            pkgSpecifier
@@ -186,11 +213,12 @@ info verbosity packageDBs repos comp conf
 
   where
     gatherPkgInfo :: (PackageName -> VersionRange) ->
-                     InstalledPackageIndex.PackageIndex ->
-                     PackageIndex.PackageIndex SourcePackage ->
-                     PackageSpecifier SourcePackage ->
+                     InstalledPackageIndex ->
+                     PackageIndex.PackageIndex UnresolvedSourcePackage ->
+                     PackageSpecifier UnresolvedSourcePackage ->
                      Either String PackageDisplayInfo
-    gatherPkgInfo prefs installedPkgIndex sourcePkgIndex (NamedPackage name constraints)
+    gatherPkgInfo prefs installedPkgIndex sourcePkgIndex
+      (NamedPackage name props)
       | null (selectedInstalledPkgs) && null (selectedSourcePkgs)
       = Left $ "There is no available version of " ++ display name
             ++ " that satisfies "
@@ -204,19 +232,21 @@ info verbosity packageDBs repos comp conf
         (pref, installedPkgs, sourcePkgs) =
           sourcePkgsInfo prefs name installedPkgIndex sourcePkgIndex
 
-        selectedInstalledPkgs = InstalledPackageIndex.lookupDependency installedPkgIndex
-                                    (Dependency name verConstraint)
-        selectedSourcePkgs    =          PackageIndex.lookupDependency sourcePkgIndex
-                                    (Dependency name verConstraint)
+        selectedInstalledPkgs = InstalledPackageIndex.lookupDependency
+                                installedPkgIndex
+                                name verConstraint
+        selectedSourcePkgs    = PackageIndex.lookupDependency sourcePkgIndex
+                                name verConstraint
         selectedSourcePkg'    = latestWithPref pref selectedSourcePkgs
 
                          -- display a specific package version if the user
                          -- supplied a non-trivial version constraint
         showPkgVersion = not (null verConstraints)
         verConstraint  = foldr intersectVersionRanges anyVersion verConstraints
-        verConstraints = [ vr | PackageConstraintVersion _ vr <- constraints ]
+        verConstraints = [ vr | PackagePropertyVersion vr <- props ]
 
-    gatherPkgInfo prefs installedPkgIndex sourcePkgIndex (SpecificSourcePackage pkg) =
+    gatherPkgInfo prefs installedPkgIndex sourcePkgIndex
+      (SpecificSourcePackage pkg) =
         Right $ mergePackageInfo pref installedPkgs sourcePkgs
                                  selectedPkg True
       where
@@ -228,15 +258,16 @@ info verbosity packageDBs repos comp conf
 sourcePkgsInfo ::
   (PackageName -> VersionRange)
   -> PackageName
-  -> InstalledPackageIndex.PackageIndex
-  -> PackageIndex.PackageIndex SourcePackage
-  -> (VersionRange, [Installed.InstalledPackageInfo], [SourcePackage])
+  -> InstalledPackageIndex
+  -> PackageIndex.PackageIndex UnresolvedSourcePackage
+  -> (VersionRange, [Installed.InstalledPackageInfo], [UnresolvedSourcePackage])
 sourcePkgsInfo prefs name installedPkgIndex sourcePkgIndex =
   (pref, installedPkgs, sourcePkgs)
   where
     pref          = prefs name
-    installedPkgs = concatMap snd (InstalledPackageIndex.lookupPackageName installedPkgIndex name)
-    sourcePkgs    =                         PackageIndex.lookupPackageName sourcePkgIndex name
+    installedPkgs = concatMap snd (InstalledPackageIndex.lookupPackageName
+                                   installedPkgIndex name)
+    sourcePkgs    = PackageIndex.lookupPackageName sourcePkgIndex name
 
 
 -- | The info that we can display for each package. It is information per
@@ -245,36 +276,41 @@ sourcePkgsInfo prefs name installedPkgIndex sourcePkgIndex =
 data PackageDisplayInfo = PackageDisplayInfo {
     pkgName           :: PackageName,
     selectedVersion   :: Maybe Version,
-    selectedSourcePkg :: Maybe SourcePackage,
+    selectedSourcePkg :: Maybe UnresolvedSourcePackage,
     installedVersions :: [Version],
     sourceVersions    :: [Version],
     preferredVersions :: VersionRange,
-    homepage          :: String,
-    bugReports        :: String,
-    sourceRepo        :: String,
-    synopsis          :: String,
-    description       :: String,
-    category          :: String,
-    license           :: License,
-    author            :: String,
-    maintainer        :: String,
+    homepage          :: ShortText,
+    bugReports        :: ShortText,
+    sourceRepo        :: String, -- TODO
+    synopsis          :: ShortText,
+    description       :: ShortText,
+    category          :: ShortText,
+    license           :: Either SPDX.License License,
+    author            :: ShortText,
+    maintainer        :: ShortText,
     dependencies      :: [ExtDependency],
     flags             :: [Flag],
     hasLib            :: Bool,
     hasExe            :: Bool,
-    executables       :: [String],
+    executables       :: [UnqualComponentName],
     modules           :: [ModuleName],
     haddockHtml       :: FilePath,
     haveTarball       :: Bool
   }
 
+-- | Covers source dependencies and installed dependencies in
+-- one type.
+data ExtDependency = SourceDependency Dependency
+                   | InstalledDependency UnitId
+
 showPackageSummaryInfo :: PackageDisplayInfo -> String
 showPackageSummaryInfo pkginfo =
   renderStyle (style {lineLength = 80, ribbonsPerLine = 1}) $
      char '*' <+> disp (pkgName pkginfo)
      $+$
      (nest 4 $ vcat [
-       maybeShow (synopsis pkginfo) "Synopsis:" reflowParagraphs
+       maybeShowST (synopsis pkginfo) "Synopsis:" reflowParagraphs
      , text "Default available version:" <+>
        case selectedSourcePkg pkginfo of
          Nothing  -> text "[ Not available from any configured repository ]"
@@ -285,24 +321,25 @@ showPackageSummaryInfo pkginfo =
              | otherwise      -> text "[ Unknown ]"
          versions             -> dispTopVersions 4
                                    (preferredVersions pkginfo) versions
-     , maybeShow (homepage pkginfo) "Homepage:" text
-     , text "License: " <+> text (display (license pkginfo))
+     , maybeShowST (homepage pkginfo) "Homepage:" text
+     , text "License: " <+> either pretty pretty (license pkginfo)
      ])
      $+$ text ""
   where
-    maybeShow [] _ _ = empty
-    maybeShow l  s f = text s <+> (f l)
+    maybeShowST l s f
+        | ShortText.null l = empty
+        | otherwise        = text s <+> f (ShortText.fromShortText l)
 
 showPackageDetailedInfo :: PackageDisplayInfo -> String
 showPackageDetailedInfo pkginfo =
   renderStyle (style {lineLength = 80, ribbonsPerLine = 1}) $
    char '*' <+> disp (pkgName pkginfo)
-            <>  maybe empty (\v -> char '-' <> disp v) (selectedVersion pkginfo)
+            Disp.<> maybe empty (\v -> char '-' Disp.<> disp v) (selectedVersion pkginfo)
             <+> text (replicate (16 - length (display (pkgName pkginfo))) ' ')
-            <>  parens pkgkind
+            Disp.<> parens pkgkind
    $+$
    (nest 4 $ vcat [
-     entry "Synopsis"      synopsis     hideIfNull     reflowParagraphs
+     entryST "Synopsis"      synopsis     hideIfNull  reflowParagraphs
    , entry "Versions available" sourceVersions
            (altText null "[ Not available from server ]")
            (dispTopVersions 9 (preferredVersions pkginfo))
@@ -310,17 +347,17 @@ showPackageDetailedInfo pkginfo =
            (altText null (if hasLib pkginfo then "[ Not installed ]"
                                             else "[ Unknown ]"))
            (dispTopVersions 4 (preferredVersions pkginfo))
-   , entry "Homepage"      homepage     orNotSpecified text
-   , entry "Bug reports"   bugReports   orNotSpecified text
-   , entry "Description"   description  hideIfNull     reflowParagraphs
-   , entry "Category"      category     hideIfNull     text
-   , entry "License"       license      alwaysShow     disp
-   , entry "Author"        author       hideIfNull     reflowLines
-   , entry "Maintainer"    maintainer   hideIfNull     reflowLines
+   , entryST "Homepage"      homepage     orNotSpecified text
+   , entryST "Bug reports"   bugReports   orNotSpecified text
+   , entryST "Description"   description  hideIfNull     reflowParagraphs
+   , entryST "Category"      category     hideIfNull     text
+   , entry "License"       license      alwaysShow     (either pretty pretty)
+   , entryST "Author"        author       hideIfNull     reflowLines
+   , entryST "Maintainer"    maintainer   hideIfNull     reflowLines
    , entry "Source repo"   sourceRepo   orNotSpecified text
-   , entry "Executables"   executables  hideIfNull     (commaSep text)
+   , entry "Executables"   executables  hideIfNull     (commaSep disp)
    , entry "Flags"         flags        hideIfNull     (commaSep dispFlag)
-   , entry "Dependencies"  dependencies hideIfNull     (commaSep disp)
+   , entry "Dependencies"  dependencies hideIfNull     (commaSep dispExtDep)
    , entry "Documentation" haddockHtml  showIfInstalled text
    , entry "Cached"        haveTarball  alwaysShow     dispYesNo
    , if not (hasLib pkginfo) then empty else
@@ -333,9 +370,11 @@ showPackageDetailedInfo pkginfo =
       Just Nothing      -> empty
       Just (Just other) -> label <+> text other
       where
-        label   = text fname <> char ':' <> padding
+        label   = text fname Disp.<> char ':' Disp.<> padding
         padding = text (replicate (13 - length fname ) ' ')
 
+    entryST fname field = entry fname (ShortText.fromShortText . field)
+
     normal      = Nothing
     hide        = Just Nothing
     replace msg = Just (Just msg)
@@ -350,10 +389,13 @@ showPackageDetailedInfo pkginfo =
     orNotSpecified = altText null "[ Not specified ]"
 
     commaSep f = Disp.fsep . Disp.punctuate (Disp.char ',') . map f
-    dispFlag f = case flagName f of FlagName n -> text n
+    dispFlag = text . unFlagName . flagName
     dispYesNo True  = text "Yes"
     dispYesNo False = text "No"
 
+    dispExtDep (SourceDependency    dep) = disp dep
+    dispExtDep (InstalledDependency dep) = disp dep
+
     isInstalled = not (null (installedVersions pkginfo))
     hasExes = length (executables pkginfo) >= 2
     --TODO: exclude non-buildable exes
@@ -386,8 +428,8 @@ reflowLines = vcat . map text . lines
 --
 mergePackageInfo :: VersionRange
                  -> [Installed.InstalledPackageInfo]
-                 -> [SourcePackage]
-                 -> Maybe SourcePackage
+                 -> [UnresolvedSourcePackage]
+                 -> Maybe UnresolvedSourcePackage
                  -> Bool
                  -> PackageDisplayInfo
 mergePackageInfo versionPref installedPkgs sourcePkgs selectedPkg showVer =
@@ -402,7 +444,7 @@ mergePackageInfo versionPref installedPkgs sourcePkgs selectedPkg showVer =
     sourceVersions    = map packageVersion sourcePkgs,
     preferredVersions = versionPref,
 
-    license      = combine Source.license       source
+    license      = combine Source.licenseRaw    source
                            Installed.license    installed,
     maintainer   = combine Source.maintainer    source
                            Installed.maintainer installed,
@@ -410,30 +452,32 @@ mergePackageInfo versionPref installedPkgs sourcePkgs selectedPkg showVer =
                            Installed.author     installed,
     homepage     = combine Source.homepage      source
                            Installed.homepage   installed,
-    bugReports   = maybe "" Source.bugReports source,
-    sourceRepo   = fromMaybe "" . join
+    bugReports   = maybe mempty Source.bugReports source,
+    sourceRepo   = fromMaybe mempty . join
                  . fmap (uncons Nothing Source.repoLocation
                        . sortBy (comparing Source.repoKind)
                        . Source.sourceRepos)
                  $ source,
                     --TODO: installed package info is missing synopsis
-    synopsis     = maybe "" Source.synopsis      source,
+    synopsis     = maybe mempty Source.synopsis      source,
     description  = combine Source.description    source
                            Installed.description installed,
     category     = combine Source.category       source
                            Installed.category    installed,
     flags        = maybe [] Source.genPackageFlags sourceGeneric,
     hasLib       = isJust installed
-                || fromMaybe False
-                   (fmap (isJust . Source.condLibrary) sourceGeneric),
-    hasExe       = fromMaybe False
-                   (fmap (not . null . Source.condExecutables) sourceGeneric),
+                || maybe False (isJust . Source.condLibrary) sourceGeneric,
+    hasExe       = maybe False (not . null . Source.condExecutables) sourceGeneric,
     executables  = map fst (maybe [] Source.condExecutables sourceGeneric),
-    modules      = combine Installed.exposedModules installed
-                           (maybe [] Source.exposedModules
-                                   . Source.library) source,
-    dependencies = combine (map (SourceDependency . simplifyDependency) . Source.buildDepends) source
-                           (map InstalledDependency . Installed.depends) installed,
+    modules      = combine (map Installed.exposedName . Installed.exposedModules)
+                           installed
+                           -- NB: only for the PUBLIC library
+                           (concatMap getListOfExposedModules . maybeToList . Source.library)
+                           source,
+    dependencies =
+      combine (map (SourceDependency . simplifyDependency)
+               . Source.allBuildDepends) source
+      (map InstalledDependency . Installed.depends) installed,
     haddockHtml  = fromMaybe "" . join
                  . fmap (listToMaybe . Installed.haddockHTMLs)
                  $ installed,
@@ -444,6 +488,10 @@ mergePackageInfo versionPref installedPkgs sourcePkgs selectedPkg showVer =
     installed :: Maybe Installed.InstalledPackageInfo
     installed = latestWithPref versionPref installedPkgs
 
+    getListOfExposedModules lib = Source.exposedModules lib
+                               ++ map Source.moduleReexportName
+                                      (Source.reexportedModules lib)
+
     sourceSelected
       | isJust selectedPkg = selectedPkg
       | otherwise          = latestWithPref versionPref sourcePkgs
@@ -482,10 +530,10 @@ latestWithPref pref pkgs = Just (maximumBy (comparing prefThenVersion) pkgs)
 -- both be empty.
 --
 mergePackages :: [Installed.InstalledPackageInfo]
-              -> [SourcePackage]
+              -> [UnresolvedSourcePackage]
               -> [( PackageName
                   , [Installed.InstalledPackageInfo]
-                  , [SourcePackage] )]
+                  , [UnresolvedSourcePackage] )]
 mergePackages installedPkgs sourcePkgs =
     map collect
   $ mergeBy (\i a -> fst i `compare` fst a)
@@ -529,13 +577,13 @@ dispTopVersions n pref vs =
 --
 interestingVersions :: (Version -> Bool) -> [Version] -> [Version]
 interestingVersions pref =
-      map ((\ns -> Version ns []) . fst) . filter snd
+      map (mkVersion . fst) . filter snd
     . concat  . Tree.levels
     . swizzleTree
-    . reorderTree (\(Node (v,_) _) -> pref (Version v []))
+    . reorderTree (\(Node (v,_) _) -> pref (mkVersion v))
     . reverseTree
     . mkTree
-    . map versionBranch
+    . map versionNumbers
 
   where
     swizzleTree = unfoldTree (spine [])
